@@ -1,11 +1,13 @@
-use crate::builder::PassBuilder;
+use crate::{builder::PassBuilder, SwcImportResolver};
 use anyhow::{bail, Context, Error};
 use dashmap::DashMap;
 use either::Either;
 use once_cell::sync::Lazy;
 use regex::Regex;
 use serde::{Deserialize, Serialize};
+use std::rc::Rc as RustRc;
 use std::{
+    cell::RefCell,
     collections::{HashMap, HashSet},
     env,
     path::{Path, PathBuf},
@@ -17,9 +19,11 @@ pub use swc_common::chain;
 use swc_common::{comments::Comments, errors::Handler, FileName, Mark, SourceMap};
 use swc_ecma_ast::{Expr, ExprStmt, ModuleItem, Stmt};
 use swc_ecma_ext_transforms::jest;
+use swc_ecma_loader::resolvers::{lru::CachingResolver, node::NodeResolver, tsc::TsConfigResolver};
 pub use swc_ecma_parser::JscTarget;
 use swc_ecma_parser::{lexer::Lexer, Parser, StringInput, Syntax, TsConfig};
-use swc_ecma_transforms::hygiene;
+use swc_ecma_transforms::modules::path::NodeImportResolver;
+use swc_ecma_transforms::{hygiene, modules::util::Scope};
 use swc_ecma_transforms::{
     modules,
     optimization::const_modules,
@@ -171,6 +175,7 @@ impl Options {
     pub fn build<'a>(
         &self,
         cm: &Arc<SourceMap>,
+        base: &FileName,
         handler: &Handler,
         is_module: bool,
         config: Option<Config>,
@@ -186,6 +191,7 @@ impl Options {
             target,
             loose,
             keep_class_names,
+            paths,
         } = config.jsc;
         let target = target.unwrap_or_default();
 
@@ -262,7 +268,13 @@ impl Options {
             })
             .fixer(!self.disable_fixer)
             .preset_env(config.env)
-            .finalize(syntax, config.module, comments);
+            .finalize(
+                paths.into_iter().collect(),
+                base,
+                syntax,
+                config.module,
+                comments,
+            );
 
         let pass = chain!(pass, Optional::new(jest::jest(), transform.hidden.jest));
 
@@ -348,6 +360,7 @@ impl Default for Rc {
                     target: Default::default(),
                     loose: false,
                     keep_class_names: false,
+                    ..Default::default()
                 },
                 module: None,
                 minify: None,
@@ -368,6 +381,7 @@ impl Default for Rc {
                     target: Default::default(),
                     loose: false,
                     keep_class_names: false,
+                    ..Default::default()
                 },
                 module: None,
                 minify: None,
@@ -388,6 +402,7 @@ impl Default for Rc {
                     target: Default::default(),
                     loose: false,
                     keep_class_names: false,
+                    ..Default::default()
                 },
                 module: None,
                 minify: None,
@@ -587,8 +602,15 @@ pub struct JscConfig {
 
     #[serde(default)]
     pub keep_class_names: bool,
+
+    #[serde(default)]
+    pub paths: Paths,
 }
 
+/// `paths` sectiob of `tsconfig.json`.
+pub type Paths = HashMap<String, Vec<String>, ahash::RandomState>;
+pub(crate) type CompiledPaths = Vec<(String, Vec<String>)>;
+
 #[derive(Debug, Clone, Serialize, Deserialize)]
 #[serde(deny_unknown_fields, rename_all = "camelCase")]
 #[serde(tag = "type")]
@@ -606,16 +628,60 @@ pub enum ModuleConfig {
 impl ModuleConfig {
     pub fn build(
         cm: Arc<SourceMap>,
+        paths: CompiledPaths,
+        base: &FileName,
         root_mark: Mark,
         config: Option<ModuleConfig>,
+        scope: RustRc<RefCell<Scope>>,
     ) -> Box<dyn swc_ecma_visit::Fold> {
+        let base = match base {
+            FileName::Real(v) => {
+                FileName::Real(v.canonicalize().unwrap_or_else(|_| v.to_path_buf()))
+            }
+            _ => base.clone(),
+        };
+
         match config {
             None | Some(ModuleConfig::Es6) => Box::new(noop()),
             Some(ModuleConfig::CommonJs(config)) => {
-                Box::new(modules::common_js::common_js(root_mark, config))
+                if paths.is_empty() {
+                    Box::new(modules::common_js::common_js(
+                        root_mark,
+                        config,
+                        Some(scope),
+                    ))
+                } else {
+                    let resolver = build_resolver(paths);
+
+                    Box::new(modules::common_js::common_js_with_resolver(
+                        resolver,
+                        base,
+                        root_mark,
+                        config,
+                        Some(scope),
+                    ))
+                }
+            }
+            Some(ModuleConfig::Umd(config)) => {
+                if paths.is_empty() {
+                    Box::new(modules::umd::umd(cm, root_mark, config))
+                } else {
+                    let resolver = build_resolver(paths);
+
+                    Box::new(modules::umd::umd_with_resolver(
+                        resolver, base, cm, root_mark, config,
+                    ))
+                }
+            }
+            Some(ModuleConfig::Amd(config)) => {
+                if paths.is_empty() {
+                    Box::new(modules::amd::amd(config))
+                } else {
+                    let resolver = build_resolver(paths);
+
+                    Box::new(modules::amd::amd_with_resolver(resolver, base, config))
+                }
             }
-            Some(ModuleConfig::Umd(config)) => Box::new(modules::umd::umd(cm, root_mark, config)),
-            Some(ModuleConfig::Amd(config)) => Box::new(modules::amd::amd(config)),
         }
     }
 }
@@ -894,3 +960,24 @@ impl Merge for ConstModulesConfig {
         *self = from.clone()
     }
 }
+
+fn build_resolver(paths: CompiledPaths) -> SwcImportResolver {
+    static CACHE: Lazy<DashMap<CompiledPaths, SwcImportResolver>> =
+        Lazy::new(|| Default::default());
+
+    if let Some(cached) = CACHE.get(&paths) {
+        return (*cached).clone();
+    }
+
+    let r = {
+        let r = TsConfigResolver::new(NodeResolver::default(), ".".into(), paths.clone());
+        let r = CachingResolver::new(40, r);
+
+        let r = NodeImportResolver::new(r);
+        Arc::new(r)
+    };
+
+    CACHE.insert(paths, r.clone());
+
+    r
+}
